import java.io.*;
import java.util.*;

/** Program to solve Nurikabe puzzles. */
public class Nurikabe implements Iterable<Coords> {

	/** Character for a filled-in cell. */
	public static final char BLACK = '#';

	/** Character for a cell known to be open. */
	public static final char OPEN = '.';

	/** Character for a cell of unknown status. */
	public static final char UNKNOWN = '?';

	/** Statistics for a solve() call */
	int backtrackCount;
	int branchCount;

	ArrayList<Coords> numberCoords;

	/** Read the specified puzzle, solve it, and verify the solution. */
	public static void main(String[] args) {
		if (args.length > 0) {
			Nurikabe puzzle = new Nurikabe(args[0] + ".puz");
			puzzle.solve();
			System.out.println(puzzle);
			System.out.println("Correct? " + puzzle.validate(args[0] + ".sol"));
		} else {
			for (int i = 1; i <= 8; i++) {
				String name = "sample" + i;
				Nurikabe puzzle = new Nurikabe(name + ".puz");
				puzzle.solve();
				System.out.print(puzzle);
				System.out.println("Correct? " + puzzle.validate(name + ".sol"));
			}
		}
	}

	/** Grid of cells. Each is BLACK, OPEN, UNKNOWN, or a digit 1-9. */
	private char[][] grid;

        /** Read a puzzle from the specified file. Replace any .s with UNKNOWN. */
	public Nurikabe(String filename) {
		try {
			Scanner in = new Scanner(new FileInputStream(filename));
			ArrayList<String> lines = new ArrayList<String>();
			while (in.hasNextLine()) {
				lines.add(in.nextLine());
			}
			grid = new char[lines.size()][lines.get(0).length()];
			for (int r = 0; r < getHeight(); r++) {
				for (int c = 0; c < getWidth(); c++) {
					char symbol = lines.get(r).charAt(c);
					grid[r][c] = symbol == OPEN ? UNKNOWN : symbol;
				}
			}
		} catch (IOException e) {
			e.printStackTrace();
			System.exit(1);
		}
	}

	/** Create a copy of the grid. Useful for undoing during search. */
	protected char[][] copyGrid() {
		char[][] result = new char[getHeight()][getWidth()];
		for (int r = 0; r < getHeight(); r++) {
			for (int c = 0; c < getWidth(); c++) {
				result[r][c] = grid[r][c];
			}
		}
		return result;
	}

	/** Return the label at a particular location. */
	public char get(Coords cell) {
		return grid[cell.getRow()][cell.getColumn()];
	}

	/** Return the number of rows in the puzzle. */
	protected int getHeight() {
		return grid.length;
	}

	/** Return the number of columns in the puzzle. */
	protected int getWidth() {
		return grid[0].length;
	}

	/** Return true if the given characters is a digit (1-9). */
	protected boolean isANumber(char c) {
		return (c >= '1') && (c <= '9');
	}

	/** Iterate through the cells in the grid. */
	public Iterator<Coords> iterator() {
		return new NurikabeIterator(this);
	}

	/** Return the coordinates of all cells adjacent to a given cell. */
	protected List<Coords> neighbors(Coords cell) {
		return neighbors(cell.getRow(), cell.getColumn());
	}

	/** Return the coordinates of all cells adjacent to a given cell. */
	protected List<Coords> neighbors(int r, int c) {
		List<Coords> result = new ArrayList<Coords>();
		if (r > 0) {
			result.add(new Coords(r - 1, c));
		}
		if (r < getHeight() - 1) {
			result.add(new Coords(r + 1, c));
		}
		if (c > 0) {
			result.add(new Coords(r, c - 1));
		}
		if (c < getWidth() - 1) {
			result.add(new Coords(r, c + 1));
		}
		return result;
	}

	/** Return the number of cells in the grid with the specified label. */
	public int numberOfCells(char label) {
		int tally = 0;
		for (Coords cell : this) {
			if (get(cell) == label) {
				tally++;
			}
		}
		return tally;
	}

	/** Set the label at a given location. */
	protected void set(Coords cell, char value) {
		grid[cell.getRow()][cell.getColumn()] = value;
	}

	public boolean isLegal(int x, int y, char value) {
		switch (value) {
		case BLACK:
			/* Check for 2x2 square creation */
			if (x > 0) {
				if (y > 0) {
					/* Upper left square */
					if (grid[y-1][x-1] == BLACK &&
					    grid[y-1][x] == BLACK &&
					    grid[y][x-1] == BLACK)
						return false;
				}
				if (y < grid.length - 1) {
					/* Lower left square */
					if (grid[y+1][x-1] == BLACK &&
					    grid[y+1][x] == BLACK &&
					    grid[y][x-1] == BLACK)
						return false;
				}
			}
			if (x < grid[y].length - 1) {
				if (y > 0) {
					/* Upper right square */
					if (grid[y-1][x+1] == BLACK &&
					    grid[y-1][x] == BLACK &&
					    grid[y][x+1] == BLACK)
						return false;
				}
				if (y < grid.length - 1) {
					/* Lower right square */
					if (grid[y+1][x+1] == BLACK &&
					    grid[y+1][x] == BLACK &&
					    grid[y][x+1] == BLACK)
						return false;
				}
			}
			return true;
		case OPEN:
			/* Groups getting connected will get caught at the end
			 */
			return true;
		default:
			return false;
		}
	}

	void debug(String msg) {
		if (false)
			System.out.print(msg);
	}

	boolean boardHas2x2() {
		for (int y = 0; y < getHeight() - 1; y++) {
		    for (int x = 0; x < getWidth() - 1; x++) {
			int blackCount = 0;
			int unknownX = 0, unknownY = 0;

			if (grid[y + 0][x + 0] == BLACK &&
			    grid[y + 0][x + 1] == BLACK &&
			    grid[y + 1][x + 0] == BLACK &&
			    grid[y + 1][x + 1] == BLACK)
				return true;
		    }
		}
		return false;
	}

	boolean isBoardSane() {
		/* Check for numbers getting enclosed too early */
		for (Coords coord : numberCoords) {
			int x = coord.getColumn();
			int y = coord.getRow();
			int number = grid[y][x] - '0';
			Region r = new Region(this, new Coords(y, x),
				".123456789");
			Region n = r.getNeighbors("?");
			if (n.size() == 0 && r.size() != number)
				return false;
		}
		
		return (!boardHas2x2());
	}

	boolean applyCornerRule(Region neighbors) {
		Coords[] coords = new Coords[2];
		int i = 0, x0, y0, x1, y1;
		boolean changed = false;

		for (Coords cell : neighbors) {
			coords[i++] = cell;
		}
		x0 = coords[0].getColumn();
		y0 = coords[0].getRow();
		x1 = coords[1].getColumn();
		y1 = coords[1].getRow();
		if ((x1 - x0 == 1 || x1 - x0 == -1) &&
		    (y1 - y0 == 1 || y1 - y0 == -1)) {
			if (grid[y0][x1] == UNKNOWN) {
				changed = true;
				grid[y0][x1] = BLACK;
				debug("corner fill rule at (" + x1 + "," + y0 +
				    ")\n");
			}
			if (grid[y1][x0] == UNKNOWN) {
				changed = true;
				grid[y1][x0] = BLACK;
				debug("corner fill rule at (" + x0 + "," + y1 +
				    ")\n");
			}
		}

		return changed;
	}

	/**
	 * If there are two liberties and the group is at least 1 short, and one
	 * liberty is a dead end, then extend the other way.
	 */
	boolean applyNumber2LibertyRule(int gy, int gx, Region n) {
		Coords libertycell = null;
		boolean multiliberty = false;

		for (Coords cell2 : n) {
			int y1 = cell2.getRow();
			int x1 = cell2.getColumn();
			int blockcount = 0;

			if (isBlocked(y1 + 1, x1) || hasNumNeighbor(gy, gx, y1 + 1, x1))
				blockcount++;
			if (isBlocked(y1, x1 + 1) || hasNumNeighbor(gy, gx, y1, x1 + 1))
				blockcount++;
			if (isBlocked(y1 - 1, x1) || hasNumNeighbor(gy, gx, y1 - 1, x1))
				blockcount++;
			if (isBlocked(y1, x1 - 1) || hasNumNeighbor(gy, gx, y1, x1 - 1))
				blockcount++;
			/* blocked up except for its connection
			 * to the existing group.
			 */
			if (blockcount == 3)
				continue;
			if (libertycell != null) {
				multiliberty = true;
				break;
			}
			libertycell = cell2;
		}
		if ((libertycell != null) && !multiliberty) {
			int y1 = libertycell.getRow();
			int x1 = libertycell.getColumn();
			grid[y1][x1] = OPEN;
			debug("number group multi-liberty " +
			    "extend at (" + x1 + "," + y1 +
			    ")\n");
			return true;
		}

		return false;
	}

	boolean applyWhiteLibertyRules() {
		boolean changed = false;

		/* Check for fulfilled or nearly-fulfilled numbers */
		for (Coords coord : numberCoords) {
			int x = coord.getColumn();
			int y = coord.getRow();
			int number = grid[y][x] - '0';
			Region r = new Region(this, new Coords(y, x),
				".123456789");
			if (r.size() == number) {
				/* This number has been satisfied, so
				 * surround it.
				 */
				Region neighbors = r.getNeighbors("?");
				for (Coords cell : neighbors) {
					int nx = cell.getColumn();
					int ny = cell.getRow();
					
					grid[ny][nx] = BLACK;
					changed = true;
					debug("surround at (" + nx +
						"," + ny + ")\n");
				}
			} else {
				/* Check if this number group has only
				 * one liberty,	and extend if so.
				 */
				Region neighbors = r.getNeighbors("?");
				if (neighbors.size() == 1) {
					Coords cell = neighbors.iterator().next();
					int nx = cell.getColumn();
					int ny = cell.getRow();
					grid[ny][nx] = OPEN;
					changed = true;
					debug("open extend at (" + nx +
						"," + ny + ")\n");
				} else if (r.size() == (number - 1) &&
				    neighbors.size() == 2) {
					changed |= applyCornerRule(neighbors);
				} else if (neighbors.size() == 2) {
					applyNumber2LibertyRule(y, x, neighbors);
				}
			}
		}

		return changed;
	}

	/**
	 * Fill in shared unknown neighbors between number groups.  Fill in
	 * shared unknown neighbor between a number group and isolated group if
	 * opening it would make a too-large group.
	 */
	boolean applySharedNeighborRule() {
		boolean changed = false;

		int[][] neighborCount = new int[getHeight()][getWidth()];
		int[][] numberNeighborCount = new int[getHeight()][getWidth()];
		Region checkedSoFar = new Region(this);

		for (Coords coord : numberCoords) {
			int x = coord.getColumn();
			int y = coord.getRow();

			Region r = new Region(this, new Coords(y, x),
			    ".123456789");
			checkedSoFar.addAll(r);
			r = r.getNeighbors("?");
			for (Coords c : r) {
				neighborCount[c.getRow()][c.getColumn()]++;
				numberNeighborCount[c.getRow()][c.getColumn()]++;
			}
		}

		for (int y = 0; y < getHeight(); y++) {
		    for (int x = 0; x < getWidth(); x++) {
			Coords coord = new Coords(y, x);
			if (checkedSoFar.contains(coord))
				continue;
			Region r = new Region(this, coord, ".");
			checkedSoFar.addAll(r);
			r = r.getNeighbors("?");
			for (Coords c : r) {
				neighborCount[c.getRow()][c.getColumn()]++;
			}
		    }
		}

		for (int y = 0; y < getHeight(); y++) {
		    for (int x = 0; x < getWidth(); x++) {
			if (numberNeighborCount[y][x] >= 2) {
				grid[y][x] = BLACK;
				changed = true;
				debug("shared neighbor rule at (" + x + "," +
				    y + ")\n");
			} else if (numberNeighborCount[y][x] == 1 &&
			    neighborCount[y][x] >= 2) {
				Coords coord = new Coords(y, x);
				Region r = new Region(this, coord, ".123456789");
				int number = 0;

				for (Coords c : r) {
					char v = grid[c.getRow()][c.getColumn()];
					if (isANumber(v)) {
						number = v - '0';
						break;
					}
				}

				if (r.size() > number) {
					debug("shared neighbor limit at (" + x +
					    "," + y + ")\n");
					grid[y][x] = BLACK;
					changed = true;
				}
			}
		    }
		}

		return changed;
	}

	/* Force open a single unknown in any 2x2 block with 3 black. */
	boolean apply2x2Rule() {
		for (int y = 0; y < getHeight() - 1; y++) {
		    for (int x = 0; x < getWidth() - 1; x++) {
			int blackCount = 0;
			int unknownX = 0, unknownY = 0;

			if (grid[y][x] == BLACK) {
				blackCount++;
			} else if (grid[y][x] == UNKNOWN) {
				unknownX = x;
				unknownY = y;
			} else {
				continue;
			}

			if (grid[y][x + 1] == BLACK) {
				blackCount++;
			} else if (grid[y][x + 1] == UNKNOWN) {
				unknownX = x + 1;
				unknownY = y;
			} else {
				continue;
			}

			if (grid[y + 1][x] == BLACK) {
				blackCount++;
			} else if (grid[y + 1][x] == UNKNOWN) {
				unknownX = x;
				unknownY = y + 1;
			} else {
				continue;
			}

			if (grid[y + 1][x + 1] == BLACK) {
				blackCount++;
			} else if (grid[y + 1][x + 1] == UNKNOWN) {
				unknownX = x + 1;
				unknownY = y + 1;
			} else {
				continue;
			}

			if (blackCount == 3) {
				grid[unknownY][unknownX] = OPEN;
				debug("forced open at (" + unknownX + "," +
				    unknownY + ")\n");
				return true;
			}
		    }
		}

		return false;
	}

	/**
	 * Check if x1,y1 or x2,y2 has only a single liberty (the other one) if
	 * set to white.  Treats existing white groups as liberties, as checking
	 * for their satisfaction would be more complex.
	 */
	private boolean innerApply2x2HalfRule(int y1, int x1, int y2, int x2) {
		Region r;
		int blockcount = 0;

		if (isBlocked(y1 + 1, x1))
			blockcount++;
		if (isBlocked(y1, x1 + 1))
			blockcount++;
		if (isBlocked(y1 - 1, x1))
			blockcount++;
		if (isBlocked(y1, x1 - 1))
			blockcount++;
		if (blockcount == 3) {
			debug("2x2 1/2 fill rule at (" + x1 + "," + y1 + ")\n");
			grid[y2][x2] = OPEN;
			return true;
		}

		blockcount = 0;
		if (isBlocked(y2 + 1, x2))
			blockcount++;
		if (isBlocked(y2, x2 + 1))
			blockcount++;
		if (isBlocked(y2 - 1, x2))
			blockcount++;
		if (isBlocked(y2, x2 - 1))
			blockcount++;
		if (blockcount == 3) {
			debug("2x2 1/2 fill rule at (" + x2 + "," + y2 + ")\n");
			grid[y1][x1] = OPEN;
			return true;
		}
		return false;
	}

	/**
	 * In any 2x2 block with two adjacent blacks and two adjacent unknowns,
	 * check if setting one of the unknowns would result in the new white
	 * group being separated and its only liberty being the other unknown.
	 */
	boolean apply2x2HalfRule() {
		boolean changed = false;

		for (int y = 0; y < getHeight() - 1; y++) {
		    for (int x = 0; x < getWidth() - 1; x++) {
			if (grid[y][x] == BLACK &&
			    grid[y][x + 1] == BLACK &&
			    grid[y + 1][x] == UNKNOWN &&
			    grid[y + 1][x + 1] == UNKNOWN) {
				changed |= innerApply2x2HalfRule(y + 1, x,
				    y + 1, x + 1);
			}
			if (grid[y + 1][x] == BLACK &&
			    grid[y + 1][x + 1] == BLACK &&
			    grid[y][x] == UNKNOWN &&
			    grid[y][x + 1] == UNKNOWN) {
				changed |= innerApply2x2HalfRule(y, x,
				    y, x + 1);
			}
			if (grid[y][x] == BLACK &&
			    grid[y + 1][x] == BLACK &&
			    grid[y][x + 1] == UNKNOWN &&
			    grid[y + 1][x + 1] == UNKNOWN) {
				changed |= innerApply2x2HalfRule(y, x + 1,
				    y + 1, x + 1);
			}
			if (grid[y][x + 1] == BLACK &&
			    grid[y + 1][x + 1] == BLACK &&
			    grid[y][x] == UNKNOWN &&
			    grid[y + 1][x] == UNKNOWN) {
				changed |= innerApply2x2HalfRule(y, x,
				    y + 1, x);
			}
		    }
		}

		return changed;
	}

	/**
	 * If an isolated white group has no liberties, it must be filled.
	 * If an isolated white group has one liberty, extend it.
	 * If an isolated white group has two (n > 1?) liberties, check if one of them
	 * is an immediate dead end, and extend the other one if so.
	 */
	boolean applyIsolatedWhiteRules() {
		boolean changed = false;

		Region checkedWhite = new Region(this);

		for (int y = 0; y < getHeight(); y++) {
		    for (int x = 0; x < getWidth(); x++) {
			boolean hasNumber = false;

			if (grid[y][x] != OPEN)
				continue;

			/* If we've already looked at this group, skip it. */
			Coords coord = new Coords(y, x);
			if (checkedWhite.contains(coord))
				continue;

			Region r = new Region(this, new Coords(y, x),
				".123456789");
			checkedWhite.addAll(r);

			/* Check for isolation from numbers */
			for (Coords cell : r) {
				int gx = cell.getColumn();
				int gy = cell.getRow();
				if (grid[gy][gx] != OPEN) {
					hasNumber = true;
					break;
				}
			}
			if (hasNumber)
				continue;

			/* Grab unknown neighbors, count liberties */
			Region n = r.getNeighbors("?");
			switch (n.size()) {
			case 0:
				/* Actually, wait, this should never happen.
				 * Right?  (except in illegal board states,
				 * which this should help kill sooner)
				 */
				for (Coords cell : r) {
					int nx = cell.getColumn();
					int ny = cell.getRow();
					grid[ny][nx] = BLACK;
					changed = true;
					debug("isolated fill at (" + nx +
						"," + ny + ")\n");
				}
				break;
			case 1:
				Coords cell = n.iterator().next();
				int nx = cell.getColumn();
				int ny = cell.getRow();
				grid[ny][nx] = OPEN;
				checkedWhite.add(new Coords(ny, nx));
				changed = true;
				debug("isolated open extend at (" + nx +
					"," + ny + ")\n");
				break;
			default:
				Coords libertycell = null;
				boolean multiliberty = false;

				for (Coords cell2 : n) {
					int y1 = cell2.getRow();
					int x1 = cell2.getColumn();
					int blockcount = 0;

					if (isBlocked(y1 + 1, x1))
						blockcount++;
					if (isBlocked(y1, x1 + 1))
						blockcount++;
					if (isBlocked(y1 - 1, x1))
						blockcount++;
					if (isBlocked(y1, x1 - 1))
						blockcount++;
					/* blocked up except for its connection
					 * to the existing group.
					 */
					if (blockcount == 3)
						continue;
					if (libertycell != null) {
						multiliberty = true;
						break;
					}
					libertycell = cell2;
				}
				if ((libertycell != null) && !multiliberty) {
					int y1 = libertycell.getRow();
					int x1 = libertycell.getColumn();
					grid[y1][x1] = OPEN;
					debug("isolated group multi-liberty " +
					    "extend at (" + x1 + "," + y1 +
					    ")\n");
				}
				break;
			}
		    }
		}

		return changed;
	}

	/**
	 * If there is more than one black group, and any black group has a
	 * single liberty, then fill that liberty.
	 */
	boolean applyBlackIsolationRule() {
		boolean changed = false;

		if (isBlackAllConnected())
			return false;

		Region checkedBlack = new Region(this);

		for (int y = 0; y < getHeight(); y++) {
		    for (int x = 0; x < getWidth(); x++) {
			if (grid[y][x] != BLACK)
				continue;

			Coords coord = new Coords(y, x);

			if (checkedBlack.contains(coord))
				continue;

			Region r = new Region(this, coord,"#");
			checkedBlack.addAll(r);

			Region neighbors = r.getNeighbors("?");
			if (neighbors.size() == 1) {
				Coords cell = neighbors.iterator().next();
				int nx = cell.getColumn();
				int ny = cell.getRow();
				grid[ny][nx] = BLACK;
				debug("black extend at (" + nx +
					"," + ny + ")\n");
				changed = true;
				if (isBlackAllConnected())
				    return true;
			}
		    }
		}

		return changed;
	}

	boolean isBlocked(int y, int x) {
		if (x < 0 || y < 0 || x >= getWidth() || y >= getHeight())
			return true;
		return (grid[y][x] == BLACK);
	}

	boolean isLiberty(int y, int x) {
		if (x < 0 || y < 0 || x >= getWidth() || y >= getHeight())
			return false;
		return (grid[y][x] == UNKNOWN);
	}

	boolean isNumberSquare(int y, int x) {
		if (x < 0 || y < 0 || x >= getWidth() || y >= getHeight())
			return false;
		return (isANumber(grid[y][x]));
	}

	/**
	 * Return true if the x, y has a number neighbor which isn't gx, gy
	 */
	boolean hasNumNeighbor(int gy, int gx, int y, int x) {
		if (y + 1 != gy && x != gx && isNumberSquare(y + 1, gx))
			return true;
		if (y != gy && x + 1!= gx && isNumberSquare(y, gx + 1))
			return true;
		if (y - 1 != gy && x != gx && isNumberSquare(y - 1, gx))
			return true;
		if (y != gy && x - 1 != gx && isNumberSquare(y, gx - 1))
			return true;

		return false;
	}

	boolean checkCornerCase(int y, int x, int blackYDir, int blackXDir) {
		int cornerX = x - blackXDir;
		int cornerY = y - blackYDir;

		if (isBlocked(y + blackYDir, x) && isBlocked(y, x + blackXDir) &&
		    isLiberty(y - blackYDir, x) && isLiberty(y, x - blackYDir)) {
			if (grid[cornerY][cornerX] == UNKNOWN) {
				grid[cornerY][cornerX] = BLACK;
				debug("corner rule fill at (" + cornerX + "," +
				    cornerY + ")\n");
				return true;
			}
		}

		return false;
	}

	/**
	 * Created before applyCornerRule(), this one checks the specific case
	 * of the number 2 with the corner rule.  Unused.
	 */
	boolean applyCornerRuleSimple() {
		boolean changed = false;

		for (Coords coord : numberCoords) {
			int x = coord.getColumn();
			int y = coord.getRow();
			int number = grid[y][x] - '0';

			if (number != 2)
				continue;

			if (x > 0 && y > 0) {
				changed |= checkCornerCase(x, y, 1, 1);
			}
			if (x > 0 && y < getHeight()) {
				changed |= checkCornerCase(x, y, 1, -1);
			}
			if (x < getWidth() && y > 0) {
				changed |= checkCornerCase(x, y, -1, 1);
			}
			if (x < getWidth() && y < getHeight()) {
				changed |= checkCornerCase(x, y, 1, 1);
			}
		}
		return changed;
	}

	/**
	 * Check for an unknown square blocked on all sides.  Those can't be
	 * white.
	 */
	boolean applySurroundedUnknownRule() {
		boolean changed = false;

		for (int y = 0; y < getHeight(); y++) {
		    for (int x = 0; x < getWidth(); x++) {
			if (grid[y][x] != UNKNOWN)
				continue;

			if (isBlocked(y + 1, x) &&
			    isBlocked(y, x + 1) &&
			    isBlocked(y - 1, x) &&
			    isBlocked(y, x - 1)) {
				debug("surrounded unknown fill at (" + x + "," +
				    y + ")\n");
				grid[y][x] = BLACK;
				changed = true;
			}
		    }
		}

		return changed;
	}

	/* Apply the logic rules we know, once, to the grid.  Return whether we
	 * made any changes in the process.
	 */
	boolean applyLogic() {
		boolean changed = false;

		changed |= applyWhiteLibertyRules();
		changed |= applySharedNeighborRule();
		changed |= apply2x2Rule();
		changed |= apply2x2HalfRule();
		changed |= applyIsolatedWhiteRules();
		changed |= applyBlackIsolationRule();
		changed |= applySurroundedUnknownRule();

		return changed;
	}

	/**
	 * Apply our logical rules until we run out of forced moves to make.
	 * Bail early if our board is in an illegal state.
	 */
	public void applyLogicUntilStumped() {
		boolean changed = true;

		/* Loop using our rules to improve the situation if we can. */
		while (changed) {
			int numCells = getHeight() * getWidth();
			debug("New pass (" +
			    (numCells - numberOfCells(UNKNOWN)) +
			    " of " + numCells + " filled)\n");
			changed = applyLogic();
			if (!isBoardSane())
				return;
		}
	}

	/**
	 * Begin by checking the board to see if there are squares that couldn't
	 * be reached by any of the numbers, and set them black.
	 */
	void applyReachableRule() {
		for (int y = 0; y < getHeight(); y++) {
		    for (int x = 0; x < getWidth(); x++) {
			boolean reachable = false;

			for (Coords c : numberCoords) {
				int gx = c.getColumn();
				int gy = c.getRow();
				int number = grid[gy][gx] - '0';

				int distance = Math.abs(gx - x) + Math.abs(gy - y);
				if (distance < number) {
					reachable = true;
					break;
				}
			}
			if (!reachable) {
				grid[y][x] = BLACK;
				debug("filling unreachable square (" + x + "," +
				    y + ")\n");
			}
		    }
		}
	}

	/**
	 * Solve the puzzle and return true if successful.
	 */
	public boolean solve() {
		/* Prepare the list of where the coordinates are, for future
		 * use.
		 */
		numberCoords = new ArrayList<Coords>();
		for (int y = 0; y < getHeight(); y++) {
		    for (int x = 0; x < getWidth(); x++) {
			if (isANumber(grid[y][x])) {
				numberCoords.add(new Coords(y, x));
			}
		    }
		}

		branchCount = 0;
		backtrackCount = 0;

		applyReachableRule();

		applyLogicUntilStumped();
		debug("Logical conclusions:\n" + toString());

		debug("Now solving with brute force...\n");

		boolean returnval = recursiveSolve();
		
		debug("Branched " + branchCount + " times, with " +
		    backtrackCount + " backtracks\n");

		return returnval;
	}

	public boolean isBlackAllConnected(){
		Region r = null;
		int count = 0;

		for (int y = 0; y < getHeight(); y++) {
		    for (int x = 0; x < getWidth(); x++) {

			if (grid[y][x] != BLACK)
				continue;

			if (r == null) {
				r = new Region(this, new Coords(y, x), "#");
			}
			count++;
		    }
		}
		return (r.size() == count);
	}

	/**
	 * Find the first unknown x,y value and return it.
	 */
	public Coords chooseNextFillSimple() {
		for (int y = 0; y < getHeight(); y++) {
		    for (int x = 0; x < getWidth(); x++) {
			if (grid[y][x] == UNKNOWN)
				return new Coords(y, x);
		    }
		}
		return new Coords(-1, -1);
	}

	/**
	 * Search for an x,y to set which will result in an inconsistent
	 * board for black or white.  If there are multiple, then choose
	 * the one that results in the largest quantity of board getting
	 * filled.
	 */
	public Coords chooseNextFillHeuristic() {
		int bestX = -1, bestY = -1;
		char[][] savegrid = grid;
		int mostInconsistencies = -1;
		int fewestUnknowns = getHeight() * getWidth();

		for (int y = 0; y < getHeight(); y++) {
		    for (int x = 0; x < getWidth(); x++) {
			int inconsistencies = 0;

			if (grid[y][x] != UNKNOWN)
				continue;

			grid = copyGrid();
			grid[y][x] = BLACK;
			applyLogicUntilStumped();
			if (!isBoardSane())
				inconsistencies++;
			grid = savegrid;
	
			grid = copyGrid();
			grid[y][x] = OPEN;
			applyLogicUntilStumped();
			if (!isBoardSane())
				inconsistencies++;
			grid = savegrid;
	
			/* If setting this square either way results in an
			 * illegal board, then this grid is unsolvable, and an
			 * excellent candidate to look at.
			 */
			if (inconsistencies == 2)
				return new Coords(y, x);
	
			if (inconsistencies > mostInconsistencies) {
				mostInconsistencies = inconsistencies;
				fewestUnknowns = numberOfCells(UNKNOWN);
				bestX = x;
				bestY = y;
			} else if (inconsistencies == mostInconsistencies) {
				int unknowns = numberOfCells(UNKNOWN);
				if (unknowns < fewestUnknowns) {
					fewestUnknowns = unknowns;
					bestX = x;
					bestY = y;
				}
			}
		    }
		}
		return new Coords(bestY, bestX);
	}

	public boolean recursiveSolve() {
		Region r;
		int bestX = -1, bestY = -1;
		Coords coord;
		char[][] savegrid = grid;

		if (false)
			coord = chooseNextFillHeuristic();
		else
			coord = chooseNextFillSimple();

		bestX = coord.getColumn();
		bestY = coord.getRow();

		if (bestX != -1) {
			/* We've found a square left to set */
			branchCount++;
			grid = copyGrid();
			grid[bestY][bestX] = BLACK;
			applyLogicUntilStumped();
			if (isBoardSane() && recursiveSolve()) {
				return true;
			}
			grid = savegrid;
			backtrackCount++;

			branchCount++;
			grid = copyGrid();
			grid[bestY][bestX] = OPEN;
			applyLogicUntilStumped();
			if (isBoardSane() && recursiveSolve()) {
				return true;
			}
			grid = savegrid;
			backtrackCount++;

			return false;
		}

		/* If we have reached here, the board is filled.  Check that it
		 * is a solution.
		 */

		if (!isBlackAllConnected()) {
			debug("failing for disconnected black");
			return false;
		}

		/* Check that the white areas each have one satisfied number */
		for (Coords c : numberCoords) {
			int x = c.getColumn();
			int y = c.getRow();
			boolean foundNumber = false;

			if (grid[y][x] == BLACK)
				continue;

			r = new Region(this, new Coords(y, x), ".123456789");
			for (Coords cell : r) {
				char entry = grid[cell.getRow()][cell.getColumn()];
				if (entry == OPEN)
					continue;
				/* we are looking at a number */
				if (foundNumber) {
					debug("failing for multi-number group\n");
					return false;
				}
				foundNumber = true;
				int number = entry - '0';
				if (r.size() != number) {
					debug("failing for too-small group\n");
					return false;
				}
			}
		}

		if (boardHas2x2()) {
			debug("failing for 2x2\n");
			return false;
		}

		return true;
	}

	public String toString() {
		StringBuilder result = new StringBuilder();
		for (char[] line : grid) {
			result.append(new String(line) + "\n");
		}
		return result.toString();
	}

	/**
	 * Return true if the current state of this puzzle is the same as the
	 * solution in the file.
	 */
	public boolean validate(String filename) {
		try {
			Scanner in = new Scanner(new FileInputStream(filename));
			for (int r = 0; r < getHeight(); r++) {
				if (!Arrays.equals(grid[r], in.nextLine().toCharArray())) {
					return false;
				}
			}
			return true;
		} catch (IOException e) {
			e.printStackTrace();
			System.exit(1);
			return false; // Not actually executed, but the compiler can't
			// tell this
		}
	}

}
